#ifndef __TRAYEM_H
#define __TRAYEM_H

#include <windows.h>

#include <list>
#include <string>

#include "..\litestep\wharfData.h"
#include "..\lsapi\lsapi.h"
#include "trayicon.h"

#ifdef __cplusplus
extern "C" {
#endif

__declspec( dllexport ) int initModule(HWND parent, HINSTANCE dll, wharfDataType* wd);
__declspec( dllexport ) int initModuleEx(HWND parent, HINSTANCE dll, LPCSTR szPath);
__declspec( dllexport ) void quitModule(HINSTANCE dllInst);

#ifdef __cplusplus
}
#endif

#ifdef _DEBUG
#define TRACE(str) OutputDebugString(str)
#else
#define TRACE(str) (void)0
#endif

#define MAXICONLEVEL 255

#define	WM_ICON_NOTIFY			WM_USER+10
#define	WM_REORDER_NOTIFY 	WM_USER+11
#define	WM_RECALC_NOTIFY 	  WM_USER+12

#ifndef LM_RESTORESYSTRAY
#define LM_RESTORESYSTRAY    9211
#endif
#ifndef LM_SENDSYSTRAY
#define	LM_SENDSYSTRAY		 9213
#endif
#ifndef LM_SYSTRAY
#define LM_SYSTRAY           9214
#endif

#define MAIN_TIMER      0
#define RECALC_TIMER    1
#define INIT_TIMER      2

typedef struct {
	HWND trayWnd;
	HWND tooltipWnd;
	HWND hWnd;
	int message;
	UINT uID;
	HICON hIcon;
	HICON hOriginalIcon;
	char szTip[256];
	int x,y;
	} trayType;

// Foward declarations of functions included in this code module:
ATOM				MyRegisterClass(HINSTANCE hInstance);
ATOM				MyRegisterNotifClass(HINSTANCE hInstance);
LRESULT CALLBACK	WndProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK	WndProcNotif(HWND, UINT, WPARAM, LPARAM);
void ReadConfigData();
void BangTrayTasksToggle(HWND caller, char* args);
void SetModuleOn();
void SetModuleOff();
void FreeConfigData();
void RemoveSep();
void AddSep();
void ReorderIcons();
void RecalcIcons();
void RefreshState();

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);


HWND FindLastTrayWindow();
HICON GetWindowIcon(HWND hwnd, BYTE &nLevel);
BOOL IsWindowSelected(HWND hwnd);


/////////////////////////////////////
//
// Types
//
//using namespace std;

struct STaskData
{
   STaskData(HWND _hWndNotif, HWND _hWnd)
     : hWnd(_hWnd), hWndNotif(_hWndNotif), nIconLevel(MAXICONLEVEL){};
   virtual ~STaskData()
   {
     if(IsWindow(hWndNotif))
     {
       DestroyWindow(hWndNotif);
     };
   };

   HWND hWnd;
   HWND hWndNotif;
   BYTE nIconLevel;
   CTrayIcon icon;
};

enum SpecialWndDataType
{
  swdCaptionType,
  swdProcessType,
  swdClassType
};

struct SSpecialWndData
{
  SSpecialWndData(SpecialWndDataType _type, char* _value)
    : type(_type), value(_value){};
  virtual ~SSpecialWndData(){};
  
  SpecialWndDataType type;
  std::string        value;
};

struct SMouseStroke
{
  bool isCtrl;
  bool isAlt;
  bool isShift;

  WORD message;
};

// Really trayed windows
// Stored only for handling NIM_MODIFY without NIM_ADD
struct STrayWndData
{
  HWND hwnd;
  UINT uID;
  STrayWndData(HWND _hwnd, UINT _uID) : hwnd(_hwnd), uID(_uID) {};
  
  bool operator==(STrayWndData t2)
  {
    return hwnd == t2.hwnd && uID == t2.uID;
  }
  
};
/*
bool operator==(STrayWndData t1, STrayWndData t2)
{
  return t1.hwnd == t2.hwnd && t1.uID == t2.uID;
};
*/

class TaskListType : public std::list<STaskData*>
{
public:
  /*
	void safe_remove(STaskData* const &p)
  {
    remove(p);
    delete p;
  };
*/
	void safe_erase(iterator i)
  {
    delete (*i);
    erase(i);
  };

  void safe_clear()
  {
    iterator i;
    for(i = begin(); i != end(); ++i)
      delete (*i);
    clear();
  };

  bool find_hwnd(HWND hwnd)
  {
    iterator i;
    for(i = end(); i != begin(); )
    {
      i--;
      if((*i)->hWnd == hwnd)
      {
        return true;
      }
      // skip separators in front of list
      if((*i)->hWnd == (*i)->hWndNotif)
      {
        return false;
      }
    };
    return false;
  };
  /*
  bool find_notif_hwnd(HWND hwnd)
  {
    iterator i;
    for(i = begin(); i != end(); ++i)
    {
      if((*i)->hWndNotif == hwnd)
      {
        return true;
      }
    };
    return false;
  };
  */
  void move_to_right()
  {
    iterator i;

    //move to right for new icons
    for(i = begin(); i != end(); ++i)
    {
      STaskData *pData = (*i);
      pData->icon.MoveToRight();
    };
  };

};


class SpecialWndListType : public std::list<SSpecialWndData*>
{
public:
  /*
	void safe_remove(SSpecialWndData* const &p)
  {
    remove(p);
    delete p;
  };
*/
	void safe_erase(iterator i)
  {
    delete (*i);
    erase(i);
  };

  void safe_clear()
  {
    iterator i;
    for(i = begin(); i != end(); ++i)
      delete (*i);
    clear();
  };

  bool find_typed_value(SpecialWndDataType type, char* src_string)
  {
    iterator i;
    for(i = begin(); i != end(); ++i)
    {
      if((*i)->type == type && NULL != strstr(src_string, (*i)->value.c_str()))
      {
        return true;
      }
    };
    return false;
  }
};


class TrayWndListType : public std::list<STrayWndData>
{
public:
  bool find(STrayWndData& t)
  {
    iterator i;
    for(i = begin(); i != end(); ++i)
    {
      if((*i) == t)
      {
        return true;
      }
    };
    return false;
  };
};

#endif
