/////////////////////////////////////////////////////////////////////////////
// TrayIcon.cpp : implementation file
//
// This is a conglomeration of ideas from the MSJ "Webster" application,
// sniffing round the online docs, and from other implementations such
// as PJ Naughter's "CTrayIconifyIcon" (http://indigo.ie/~pjn/ntray.html)
// especially the "CTrayIcon::OnTrayNotification" member function.
//
// This class is a light wrapper around the windows system tray stuff. It
// adds an icon to the system tray with the specified ToolTip text and 
// callback notification value, which is sent back to the Parent window.
//
// The tray icon can be instantiated using either the constructor or by
// declaring the object and creating (and displaying) it later on in the
// program. eg.
//
//		CTrayIcon m_TrayIcon;	// Member variable of some class
//		
//		... 
//		// in some member function maybe...
//		m_TrayIcon.Create(pParentWnd, WM_MY_NOTIFY, "Click here", 
//						  hIcon, nTrayIconID);
//
//  Written by Chris Maunder (Chris.Maunder@cbr.clw.csiro.au)
//
//  Adapted by Sergey Trofimov
// 
//
/////////////////////////////////////////////////////////////////////////////
#include <windows.h>

#include <windef.h>

#include <stdio.h>

#include "TrayIcon.h"

#include "trayem.h"

#include <crtdbg.h>

#include <tchar.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#ifndef AFXAPI
	#define AFXAPI __stdcall
#endif

#ifndef AFX_CDECL
	#define AFX_CDECL __cdecl
#endif

#ifndef AfxDebugBreak
#ifdef _AFX_NO_DEBUG_CRT
// by default, debug break is asm int 3, or a call to DebugBreak, or nothing
#if defined(_M_IX86) && !defined(_AFX_PORTABLE)
#define AfxDebugBreak() _asm { int 3 }
#else
#define AfxDebugBreak() DebugBreak()
#endif
#else
#define AfxDebugBreak() _CrtDbgBreak()
#endif
#endif

#ifdef _DEBUG

BOOL AFXAPI AfxAssertFailedLine(LPCSTR lpszFileName, int nLine);

#pragma comment (lib, "nafxcwd.lib")

void AFX_CDECL AfxTrace(LPCTSTR lpszFormat, ...);
// Note: file names are still ANSI strings (filenames rarely need UNICODE)
//void AFXAPI AfxAssertValidObject(const CObject* pOb,
//				LPCSTR lpszFileName, int nLine);
//void AFXAPI AfxDump(const CObject* pOb); // Dump an object from CodeView

#define TRACE              ::AfxTrace
#define THIS_FILE          __FILE__
#define ASSERT(f) \
	do \
	{ \
	if (!(f) && AfxAssertFailedLine(THIS_FILE, __LINE__)) \
		AfxDebugBreak(); \
	} while (0) \

#define VERIFY(f)          ASSERT(f)
#define ASSERT_VALID(pOb)  (::AfxAssertValidObject(pOb, THIS_FILE, __LINE__))
#define DEBUG_ONLY(f)      (f)

// The following trace macros are provided for backward compatiblity
//  (they also take a fixed number of parameters which provides
//   some amount of extra error checking)
#define TRACE0(sz)              ::AfxTrace(_T("%s"), _T(sz))
#define TRACE1(sz, p1)          ::AfxTrace(_T(sz), p1)
#define TRACE2(sz, p1, p2)      ::AfxTrace(_T(sz), p1, p2)
#define TRACE3(sz, p1, p2, p3)  ::AfxTrace(_T(sz), p1, p2, p3)

// These AFX_DUMP macros also provided for backward compatibility
#define AFX_DUMP0(dc, sz)   dc << _T(sz)
#define AFX_DUMP1(dc, sz, p1) dc << _T(sz) << p1

#else   // _DEBUG

#define ASSERT(f)          ((void)0)
#define VERIFY(f)          ((void)(f))
#define ASSERT_VALID(pOb)  ((void)0)
#define DEBUG_ONLY(f)      ((void)0)
inline void AFX_CDECL AfxTrace(LPCTSTR, ...) { }
#define TRACE              1 ? (void)0 : ::AfxTrace
#define TRACE0(sz)
#define TRACE1(sz, p1)
#define TRACE2(sz, p1, p2)
#define TRACE3(sz, p1, p2, p3)

#endif // !_DEBUG

//IMPLEMENT_DYNAMIC(CTrayIcon, CObject)

/////////////////////////////////////////////////////////////////////////////
// CTrayIcon construction/creation/destruction

CTrayIcon::CTrayIcon()
{
	memset(&m_tnd, 0, sizeof(m_tnd));
	m_bEnabled = FALSE;
	m_bHidden  = FALSE;
}

CTrayIcon::CTrayIcon(HWND hWnd, UINT uCallbackMessage, LPCTSTR szToolTip, 
					 HICON icon, UINT uID)
{
	Create(hWnd, uCallbackMessage, szToolTip, icon, uID);
	m_bHidden = FALSE;
}

BOOL CTrayIcon::Create(HWND hWnd, UINT uCallbackMessage, LPCTSTR szToolTip, 
					   HICON icon, UINT uID)
{
	// this is only for Windows 95 (or higher)
	VERIFY(m_bEnabled = ( GetVersion() & 0xff ) >= 4);
	if (!m_bEnabled) return FALSE;

	//Make sure Notification window is valid
	VERIFY(m_bEnabled = (hWnd && ::IsWindow(hWnd)));
	if (!m_bEnabled) return FALSE;
	
	//Make sure we avoid conflict with other messages
	ASSERT(uCallbackMessage >= WM_USER);

	//Tray only supports tooltip text up to 64 characters
	ASSERT(_tcslen(szToolTip) <= 64);

	// load up the NOTIFYICONDATA structure
	m_tnd.cbSize = sizeof(NOTIFYICONDATA);
	m_tnd.hWnd	 = hWnd;
	m_tnd.uID	 = uID;
	m_tnd.hIcon  = icon;
	m_tnd.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
	m_tnd.uCallbackMessage = uCallbackMessage;
	strcpy (m_tnd.szTip, szToolTip);

	// Set the tray icon
	VERIFY(m_bEnabled = Shell_NotifyIcon(NIM_ADD, &m_tnd));

	return m_bEnabled;
}

CTrayIcon::~CTrayIcon()
{
	RemoveIcon();
}


/////////////////////////////////////////////////////////////////////////////
// CTrayIcon icon manipulation

void CTrayIcon::MoveToRight()
{
	HideIcon();
	ShowIcon();
}

void CTrayIcon::RemoveIcon()
{
	if (!m_bEnabled) return;

	m_tnd.uFlags = 0;
    Shell_NotifyIcon(NIM_DELETE, &m_tnd);
    m_bEnabled = FALSE;
}

void CTrayIcon::HideIcon()
{
	if (m_bEnabled && !m_bHidden) {
		m_tnd.uFlags = NIF_ICON;
		Shell_NotifyIcon (NIM_DELETE, &m_tnd);
		m_bHidden = TRUE;
	}
}

void CTrayIcon::ShowIcon()
{
	if (m_bEnabled && m_bHidden) {
		m_tnd.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
		Shell_NotifyIcon(NIM_ADD, &m_tnd);
		m_bHidden = FALSE;
	}
}

BOOL CTrayIcon::SetIcon(HICON hIcon)
{
	if (!m_bEnabled || !hIcon) return FALSE;

	m_tnd.uFlags = NIF_ICON;
	m_tnd.hIcon = hIcon;

	return Shell_NotifyIcon(NIM_MODIFY, &m_tnd);
}

extern HINSTANCE hModule;								// current instance

BOOL CTrayIcon::SetIcon(LPCTSTR lpszIconName)
{
	HICON hIcon = LoadIcon(hModule, lpszIconName);

	return SetIcon(hIcon);
}

BOOL CTrayIcon::SetIcon(UINT nIDResource)
{
	HICON hIcon = LoadIcon(hModule, MAKEINTRESOURCE(nIDResource));

	return SetIcon(hIcon);
}

BOOL CTrayIcon::SetStandardIcon(LPCTSTR lpIconName)
{
	HICON hIcon = LoadIcon(NULL, lpIconName);

	return SetIcon(hIcon);
}

BOOL CTrayIcon::SetStandardIcon(UINT nIDResource)
{
	HICON hIcon = LoadIcon(NULL, MAKEINTRESOURCE(nIDResource));

	return SetIcon(hIcon);
}
 
HICON CTrayIcon::GetIcon() const
{
	HICON hIcon = NULL;
	if (m_bEnabled)
		hIcon = m_tnd.hIcon;

	return hIcon;
}

/////////////////////////////////////////////////////////////////////////////
// CTrayIcon tooltip text manipulation

BOOL CTrayIcon::SetTooltipText(LPCTSTR pszTip)
{
	if (!m_bEnabled) return FALSE;

	m_tnd.uFlags = NIF_TIP;
  ZeroMemory(m_tnd.szTip, 64);
	strncpy(m_tnd.szTip, pszTip, 64);
                                    
	return Shell_NotifyIcon(NIM_MODIFY, &m_tnd);
}

BOOL CTrayIcon::SetTooltipText(UINT nID)
{
	TCHAR strText[64];
	LoadString(hModule, nID, (LPTSTR) strText, 64);

	return SetTooltipText(strText);
}

LPTSTR CTrayIcon::GetTooltipText(LPTSTR buf, UINT len) const
//CString CTrayIcon::GetTooltipText() const
{
	if (m_bEnabled)
      strncpy(buf, m_tnd.szTip, len); 

	return buf;
}

/////////////////////////////////////////////////////////////////////////////
// CTrayIcon notification window stuff

BOOL CTrayIcon::SetNotificationWnd(HWND hWnd)
{
	if (!m_bEnabled) return FALSE;

	//Make sure Notification window is valid
	ASSERT(hWnd && ::IsWindow(hWnd));

	m_tnd.hWnd = hWnd;
	m_tnd.uFlags = 0;

	return Shell_NotifyIcon(NIM_MODIFY, &m_tnd);
}

HWND CTrayIcon::GetNotificationWnd() const
{
	return (m_tnd.hWnd);
}

/////////////////////////////////////////////////////////////////////////////
// CTrayIcon implentation of OnTrayNotification

extern SMouseStroke msCloseOn;
extern SMouseStroke msMinimizeOn;
extern SMouseStroke msUnMaximizeOn;
extern HWND hWndParent, hWndMe;
extern BOOL bSwitchAlways, bAutoMinimize, bHighlight;
extern HICON hCurSelIcon;

LRESULT CTrayIcon::OnTrayNotification(UINT wParam, LONG lParam) 
{
  WORD message = LOWORD(lParam);

  //Return quickly if its not for this tray icon
  if (wParam != m_tnd.uID)
    return 0L;

  if(m_tnd.hWnd == (HWND)m_tnd.uID)  // separator
  {
    if(message == WM_RBUTTONUP)
    {
      POINT pt;
      GetCursorPos(&pt);
      PostMessage( hWndParent, LM_POPUP, (WPARAM) pt.x, (LPARAM) pt.y );
    }
    else if(message == WM_LBUTTONDOWN)
    {
      PostMessage( hWndParent, LM_HIDEPOPUP, 0, 0 );
    };
    return 0;
  };

  bool isCtrl =  !!(GetKeyState(VK_CONTROL) & 0x8000);
  bool isShift = !!(GetKeyState(VK_SHIFT) & 0x8000);
  bool isAlt =   !!(GetKeyState(VK_MENU) & 0x8000);
  /*
  if(isCtrl)
    OutputDebugString(" isCtrl ");
  if(isShift)
    OutputDebugString(" isShift ");
  if(isAlt)
    OutputDebugString(" isAlt ");
  OutputDebugString("\n");
  */
  // MouseStrokes
  // 1. close window
  if(isCtrl  == msCloseOn.isCtrl  && 
     isAlt   == msCloseOn.isAlt   && 
     isShift == msCloseOn.isShift && 
     message == msCloseOn.message)
  {
    PostMessage((HWND)m_tnd.uID, WM_SYSCOMMAND, SC_CLOSE, 0L);
    RefreshState();
    return 0;
  };
  
  // 2. Minimize window
  if(isCtrl  == msMinimizeOn.isCtrl  && 
     isAlt   == msMinimizeOn.isAlt   && 
     isShift == msMinimizeOn.isShift && 
     message == msMinimizeOn.message)
  {
    PostMessage((HWND)m_tnd.uID, WM_SYSCOMMAND, SC_MINIMIZE, 0L);
    RefreshState();
    return 0;
  };
  
  // 3. (Un)Maximize window
  if(isCtrl  == msUnMaximizeOn.isCtrl  && 
     isAlt   == msUnMaximizeOn.isAlt   && 
     isShift == msUnMaximizeOn.isShift && 
     message == msUnMaximizeOn.message)
  {
    LONG style = GetWindowLong((HWND)m_tnd.uID, GWL_STYLE);
    if(style & WS_MAXIMIZEBOX)
    {
      if(IsZoomed((HWND)m_tnd.uID))
        PostMessage((HWND)m_tnd.uID, WM_SYSCOMMAND, SC_RESTORE, 0L);
      else
        PostMessage((HWND)m_tnd.uID, WM_SYSCOMMAND, SC_MAXIMIZE, 0L);
      return 0;
    };
  };
  
  
  // Standard handling
  switch(message)
  {
  case WM_MOUSEFIRST:
    {
      CHAR *szTip[256];
      if(isCtrl && isShift && isAlt)
      {
        CHAR *szClass[100];
        GetClassName((HWND)m_tnd.uID, (LPTSTR)szClass, 100);
        sprintf((LPTSTR)szTip, "Class: %s", szClass);
      }
      else
        GetWindowText((HWND)m_tnd.uID, (LPTSTR)szTip, 256);
      
      if(0 != strcmp(m_tnd.szTip, (LPCTSTR)szTip))
        SetTooltipText((LPTSTR)szTip);
    }
    break;
    /*
  case WM_LBUTTONDBLCLK:
    {
      LONG style = GetWindowLong((HWND)m_tnd.uID, GWL_STYLE);
      if(style & WS_MAXIMIZEBOX)
      {
        if(IsZoomed((HWND)m_tnd.uID))
          PostMessage((HWND)m_tnd.uID, WM_SYSCOMMAND, SC_RESTORE, 0L);
        else
          PostMessage((HWND)m_tnd.uID, WM_SYSCOMMAND, SC_MAXIMIZE, 0L);
      };
    }
    break;
    */
  case WM_RBUTTONUP:
    {
      //::SetForegroundWindow((HWND)m_tnd.uID);
      //::PostMessage(hWndMe, WM_TIMER, MAIN_TIMER, 0); 
      ::PostMessage((HWND)m_tnd.uID, 0x313, 0, GetMessagePos());
      return 0;
    };
    break;

  case WM_RBUTTONDOWN:
    ::SetForegroundWindow((HWND)m_tnd.uID);
    RefreshState();
    
    //::PostMessage((HWND)m_tnd.uID, WM_NULL, 0, 0);
    /*
		{
			BOOL result;
			HMENU hMenu;

      hMenu = GetSystemMenu((HWND)m_tnd.uID, FALSE);
      if (!hMenu) return 0;

      POINT pos;
      GetCursorPos(&pos);

      ::SetForegroundWindow((HWND)m_tnd.uID);

			result = TrackPopupMenu(hMenu, TPM_RETURNCMD|TPM_LEFTBUTTON|TPM_RIGHTBUTTON, pos.x, pos.y, 0, (HWND)m_tnd.hWnd, NULL);
			if (result) {
				PostMessage((HWND)m_tnd.uID, WM_SYSCOMMAND, (WPARAM)result, MAKELPARAM(pos.x, pos.y));
  			//DestroyMenu(hMenu);
			}

      // BUGFIX: See "PRB: Menus for Notification Icons Don't Work Correctly"
      ::PostMessage((HWND)m_tnd.hWnd, WM_NULL, 0, 0);
    }
    */
    return 0;
    break;

  case WM_LBUTTONDOWN:
    {
      // Auto minimize on click for foreground app
      if(bHighlight && bAutoMinimize && m_tnd.hIcon == hCurSelIcon)
      {
        PostMessage((HWND)m_tnd.uID, WM_SYSCOMMAND, SC_MINIMIZE, 0L);
        RefreshState();
        return 0;
      }

      SetForegroundWindow((HWND)m_tnd.uID);
      PostMessage(hWndMe, WM_TIMER, MAIN_TIMER, 0); 
      
      if(IsIconic((HWND)m_tnd.uID))
      {
        PostMessage((HWND)m_tnd.uID, WM_SYSCOMMAND, SC_RESTORE, 0L);
      };

      if(bSwitchAlways)
      {
   			SendMessage(hWndParent, LM_BRINGTOFRONT, 0, (LPARAM)m_tnd.uID);
      };
    };
    break;
  
  default:
    break;
  };
  
  return 0;
}
